#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)

#include <LiDIA:bigint.h>
#include <LiDIA:bigmod.h>
#include <LiDIA:bigrational.h>
#include <LiDIA:bigfloat.h>
#include <LiDIA:bigcomplex.h>

#include <LiDIA:base_vector.h>
#include <LiDIA:base_vector.c>


#include <LiDIA:Fp_polynomial.h>
#include <LiDIA:Fp_polynomial_util.h>

#else

#include <LiDIA/bigint.h>
#include <LiDIA/bigmod.h>
#include <LiDIA/bigrational.h>
#include <LiDIA/bigfloat.h>
#include <LiDIA/bigcomplex.h>

#include <LiDIA/base_vector.h>
#include <LiDIA/base_vector.c>


#include <LiDIA/Fp_polynomial.h>
#include <LiDIA/Fp_polynomial_util.h>

#endif


template class base_vector < TYPE >;

