@echo off
del lidiaidx.org
copy %1 lidiaidx.org
sed "s/\\unhbox\ \\voidb@x\ \\kern\ \.06em\ \\vbox\ {\\hrule\ width\.3em}/\\_/g" lidiaidx.org > lidiaidx.tm1
sed "s/\\unhbox\ \\voidb@x\ \\hbox\ {\$<\$}T\\unhbox\ \\voidb@x\ \\hbox {\$>\$}/<T>/g" lidiaidx.tm1 > lidiaidx.tm2
sed "s/\\mbox {\$\\sim \$}/\\\~{}/g" lidiaidx.tm2 > lidiaidx.tm3
sed "s/\\unhbox\ \\voidb@x\ \\hbox\ //g" lidiaidx.tm3 > lidiaidx.tm4
sed "s/\![^\.]*\./\!/g" lidiaidx.tm4 > lidiaidx.tm5
sed "s/dense\\_power\\_serie\ \\&\ //g" lidiaidx.tm5 > lidiaidx.tm6
sed "s/\\mbox {\$\\unhbox \\voidb@x \\hbox {\$\\sim \$}\$}/\\\~{}/g" lidiaidx.tm6 > lidiaidx.tm7
sed "s/\\mbox {\${\$\\sim \$}\$}/\\\~{}/g" lidiaidx.tm7 > lidiaman.idx

del lidiaidx.org
del lidiaidx.tm1
del lidiaidx.tm2
del lidiaidx.tm3
del lidiaidx.tm4
del lidiaidx.tm5
del lidiaidx.tm6
del lidiaidx.tm7

