//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995 by the LiDIA Group
//
// File        : is_prime_appl.c 
// Author      : Andreas M"uller (AM)
// Last change : AM, Feb 7 1995, initial version
//

#include <LiDIA/bigint.h>

main()
{
   bigint n;
   int a;

   cout << "\nEnter the number to test for primality: ";
   cin >> n;

   cout << "\nHow many tests shall I do? ";
   cin >> a;

   if (is_prime(n,a))
      cout <<"\nis_prime(" << n << ", " << a << ") = true\n";
   else
      cout <<"\nis_prime(" << n << ", " << a << ") = false\n";
}
   
