


#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)

#if defined (BASE_SPARSE_POWSER)
#include <LiDIA:base_sparse_power_serie.h>
#include <LiDIA:base_sparse_power_serie.c>
#endif

#if defined (BASE_DENSE_POWSER)
#include <LiDIA:base_dense_power_serie.h>
#include <LiDIA:base_dense_power_serie.c>
#endif

#if defined (COEFF_SPARSE_POWSER)
#include <LiDIA:coeff_sparse_power_serie.h>
#include <LiDIA:coeff_sparse_power_serie.c>
#endif

#if defined (SPARSE_POWSER)
#define NO_PSR_BIGMOD
#include <LiDIA:coeff_sparse_power_serie.h>
#include <LiDIA:coeff_sparse_power_serie.c>
#include <LiDIA:base_sparse_power_serie.h>
#include <LiDIA:base_sparse_power_serie.c>
#include <LiDIA:sparse_power_serie.h>
#include <LiDIA:sparse_power_serie.c>
#endif

#if defined (DENSE_POWSER)
#define NO_PSR_BIGMOD
#include <LiDIA:base_dense_power_serie.h>
#include <LiDIA:base_dense_power_serie.c>
#include <LiDIA:dense_power_serie.h>
#include <LiDIA:dense_power_serie.c>
#endif

#else

#if defined (BASE_SPARSE_POWSER)
#include <LiDIA/base_sparse_power_serie.h>
#include <LiDIA/base_sparse_power_serie.c>
#endif

#if defined (BASE_DENSE_POWSER)
#include <LiDIA/base_dense_power_serie.h>
#include <LiDIA/base_dense_power_serie.c>
#endif

#if defined (COEFF_SPARSE_POWSER)
#include <LiDIA/coeff_sparse_power_serie.h>
#include <LiDIA/coeff_sparse_power_serie.c>
#endif

#if defined (SPARSE_POWSER)
#include <LiDIA/coeff_sparse_power_serie.h>
#include <LiDIA/coeff_sparse_power_serie.c>
#include <LiDIA/base_sparse_power_serie.h>
#include <LiDIA/base_sparse_power_serie.c>
#include <LiDIA/sparse_power_serie.h>
#include <LiDIA/sparse_power_serie.c>
#endif

#if defined (DENSE_POWSER)
#include <LiDIA/base_dense_power_serie.h>
#include <LiDIA/base_dense_power_serie.c>
#include <LiDIA/dense_power_serie.h>
#include <LiDIA/dense_power_serie.c>
#endif

#endif

#if defined (BASE_SPARSE_POWSER)
template class base_sparse_power_serie < TYPE >;
#endif

#if defined (BASE_DENSE_POWSER)
template class base_dense_power_serie < TYPE >;
#endif

#if defined (COEFF_SPARSE_POWSER)
template class spc < TYPE >;
#endif

#if defined (SPARSE_POWSER)
template class                     spc< TYPE >;
template class base_sparse_power_serie< TYPE >;
template class      sparse_power_serie< TYPE >;
#endif

#if defined (DENSE_POWSER)
template class base_dense_power_serie< TYPE >;
template class      dense_power_serie< TYPE >;
#endif



