//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995 by the LiDIA Group
//
// File        : vector_error_msg.c
// Author      : Patrick Theobald (PT)
// Last change : PT, Oct 28 1995, initial version 
//               

/*
$Id: vector_error_msg.c,v 1.1 1996/12/18 15:36:35 theobald Exp $
*/

#ifndef LIDIA_POINTER
#define LIDIA_POINTER

const char *PRT = "(pointer)";

#endif

#ifndef LIDIA_VECTOR_ERROR_MSG_C
#define LIDIA_VECTOR_ERROR_MSG_C

const char *vector_error_msg[] =
{
  "(0) Parameter have to be greater than zero !",    
  "(1) Error in parameter !",                        
  "(2) Wrong vector format !!",                    
  "(3) Parameter out of range !!",                   
  "(4) Error in dimensions of the matrices !",       
  "(5) Error in format !",
  "(6) Number of allocated elements have to be greater than zero.",
  "(7) Vector can not expanded. Please check mode !!",
  "(8) Ration have to be greater than one.",
  "(9) Vectormode have to be EXPAND or FIXED.",
  "(10) Invalid sort direction !",
  "(11) Invalid cut !",
  "(12) Invalid position or number of elements !",
  "(13) Invalid cut or empty vector !",
  "(14) Differnt sizes !"
};

#endif
