rem @echo off
move config\CONFIG config\CONFIG.unix
copy config\CONFIG.os2 config\CONFIG
move makefile makefile.unix
copy makefile.os2 makefile
del config\config.h
move config\configh.os2 config\config.h
copy config\config.h include\LiDIA\config.h
echo Successfully installed the files for OS/2.
