#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)
#include <LiDIA:hash_table.h>
#include <LiDIA:hash_table.c>
#include <LiDIA:indexed_hash_table.h>
#include <LiDIA:indexed_hash_table.c>
#else
#include <LiDIA/hash_table.h>
#include <LiDIA/hash_table.c>
#include <LiDIA/indexed_hash_table.h>
#include <LiDIA/indexed_hash_table.c>
#endif

#if defined (HASH_TABLE)
template class hentry < TYPE >;
template class hash_table < TYPE >;
#endif

#if defined (I_HASH_TABLE)
template class hentry < TYPE >;
template class hash_table < TYPE >;
template class indexed_hash_table < TYPE >;
#endif
