//
// LiDIA - a library for computational number theory
//   Copyright (c) 1995 by the LiDIA Group
//
// File        : version.h
// Author      : Thomas Papanikolaou (TP)
// Last change : TP, Oct 18 1995, initial version
//

#ifndef LIDIA_VERSION_H
#define LIDIA_VERSION_H

#define LIDIA_MAJOR_VERSION 1
#define LIDIA_MINOR_VERSION 2
#define LIDIA_PATCH_VERSION 0

#endif
