//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995 by the LiDIA Group
//
// File        : system.inl
// Author      : Patrick Theobald (PT)
// Last change : PT, Sep 28 1995, initial version (1.0)
//

template < class T >
inline T *allocate(T a, lidia_size_t size)
{ 
  return new T[size];
}





