#include "lidia_elf.h"

	.text
	.align 4
	.globl USCORE(DigitVecCsubto)
USCORE(DigitVecCsubto):
	pushl %ebx
	pushl %edi
	pushl %esi
	movl 16(%esp),%eax
	movl 20(%esp),%ebx
	movl 24(%esp),%ecx
	xorl %edi,%edi
	xorl %esi,%esi
	testl %ecx,%ecx
	jz L3
L1:
	shrl $1,%esi
	movl (%eax,%edi,4),%edx
	sbbl (%ebx,%edi,4),%edx
	movl %edx,(%eax,%edi,4)
	adcl $0,%esi
	incl %edi
	decl %ecx
	jnz L1
	testl %esi,%esi
	jz L3
L2:
	subl $1,(%eax,%edi,4)
	incl %edi
	jb L2
L3: 
	popl %esi
	popl %edi
	popl %ebx
	ret
