#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)
#include <LiDIA:rational_factorization.h>
#else
#include <LiDIA/rational_factorization.h>
#endif

main()
{
  rational_factorization f, g, h;
  bigint n;
  
  cout << "\nPlease enter a number "; cin >> n;

  f.assign(n);
  f.verbose(0);
  g = f;



  f.factor();
  
  cout << "\nSign of the factorization : " << f.sign() << "\n";
  cout << "Length of the factorization : " << f.no_of_comp() << "\n";
  cout << "Factorization : " << f ;

  if (f.is_prime_factorization())
    cout<<"  is prime factorization\n";
  else 
    cout<<"  is no prime factorization\n";


  cout<<"\n\n Original factorization was "<<g;

  cout<<"\n\n Comparing the two factorizations: ";

  if (f == g)
    cout<<" equal\n";
  else cout<<" different\n";
}

