/* Funktion zur Zeitmessung, liefert 1/100 sec Ticks der user-Zeit.
	und eine Funktion, die eine Saat fuer den Zufallsgenerator
	liefert.

	RD, 12.09.90	*/

#ifndef TIMING_H
#define TIMING_H

#ifndef C_VARIETIES_H

#ifndef EXTERN_FUNCTION
#ifdef __cplusplus
#define EXTERN_FUNCTION( rtn, args ) extern "C" { rtn args; }
#define _VOID_ /* anachronism */
#else
#ifdef __STDC__
#ifdef __PARC__
#define EXTERN_FUNCTION( rtn, args ) rtn()
#define _VOID_
#else
#define EXTERN_FUNCTION( rtn, args ) rtn args
#define _VOID_ void
#endif
#else
#define EXTERN_FUNCTION( rtn, args ) rtn()
#define _VOID_
#endif
#endif
#endif

#endif

EXTERN_FUNCTION( unsigned long timeticks, (void) );
EXTERN_FUNCTION( unsigned int timeseed, (void) );

#endif
