//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995 by the LiDIA Group
//
// File        : power_test_appl.c 
// Author      : Thomas Papanikolaou (TP)
// Last change : TP, Feb 7 1995, initial version
//

#include <LiDIA/bigint.h>

int 
main()
{
  bigint a, b;
  long i;

  cout << "\n enter an bigint a: ";
  cin >> a;
  if ((i = power_test(b, a)) >= 0)
    cout << "\n input's " << i << "th root is " << b << "\n" << flush;
  else
    cout << "\n could not detect any power \n" << flush;
}
