#ifndef LIDIA_ELF_H
#define LIDIA_ELF_H

#ifdef SOLARIS
#define HAVE_ELF
#endif

#ifdef HAVE_ELF
#define USCORE(name) name
#else
#if __STDC__
#define USCORE(name) _##name
#else
#define USCORE(name) _/**/name
#endif
#endif

#endif
