#ifndef LIDIA_RANDOM_H
#define LIDIA_RANDOM_H

#include <LiDIA/lidia.h>
#include <LiDIA/iint.h>

#if !defined(HAVE_RANDOM)
EXTERN(void srandom, ( unsigned x ));
EXTERN(char * initstate, ( unsigned seed, char *arg_state, int n ));
EXTERN(char * setstate, ( char *arg_state ));
EXTERN(long random, (_VOID_));
#endif

#endif
