//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995 by the LiDIA Group
//
// File        : matrix_error_msg.c
// Author      : Patrick Theobald (PT)
// Last change : PT, Oct 28 1995, initial version 
//               

/*
$Id: matrix_error_msg.c,v 1.1 1996/12/18 15:35:15 theobald Exp $
*/

#ifndef LIDIA_MATRIX_ERROR_MSG_C
#define LIDIA_MATRIX_ERROR_MSG_C

const char *matrix_error_msg[] =
{
  "(0) Parameters have to be greater than zero !",    
  "(1) Error found in parameters !",                        
  "(2) Unknown matrix format !!",                    
  "(3) Parameters out of range !!",                   
  "(4) Error in dimensions of the matrices !",       
  "(5) Error in format !",
  "(6) No memory allocated for parameters !",
  "(7) Non square matrix !",
  "(8) No solution !!",
  "(9) File LIDIA_PRIMES not found. Please check path and filename.",
  "(10) Number of rows is not equal rank of matrix !",
  "(11) Matrix is not regular !"
};

#endif













