#include <LiDIA/bigint_lattice.h>
#include <LiDIA/bigfloat_lattice.h>
#include <LiDIA/lattice_kernel.c>

#ifndef FUNCTION
template class lll_kernel_op<E_TYPE, A_TYPE, VECTOR<E_TYPE, A_TYPE>,
                             MODULES<E_TYPE, A_TYPE, VAR> >;
#else
template class lll_kernel_fu<E_TYPE, A_TYPE, VECTOR<E_TYPE, A_TYPE>,
                             MODULES<E_TYPE, A_TYPE, VAR> >;
#endif
