//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995, 1996 by the LiDIA Group
//
// File        : port_dummy.c
// Author      : Markus Maurer (MM)
// Last change : MM, Nov 21 1996, initial version
//
// Purpose:      The only purpose of this file is to build at least
//		 one ".o"-file in this directory, so that
//		 the "ar"-command in the makefile does not exit 
//	         with an error, because no ".o"-file was found.
//               Otherwise, some "make"-programs will interrupt
//		 the make-process, if no ".o" file is found.
//

int lidia_src_port_dummy ( int i )
{
  return (i+1);
}
