#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)
#include <LiDIA:base_vector.h>
#include <LiDIA:base_vector.c>
#include <LiDIA:file_vector.h>
#include <LiDIA:file_vector.c>
#include <LiDIA:math_vector.h>
#include <LiDIA:math_vector.c>
#include <LiDIA:sort_vector.h>
#include <LiDIA:sort_vector.c>
#include <LiDIA:lidia_vector.h>
#else
#include <LiDIA/base_vector.h>
#include <LiDIA/base_vector.c>
#include <LiDIA/file_vector.h>
#include <LiDIA/file_vector.c>
#include <LiDIA/math_vector.h>
#include <LiDIA/math_vector.c>
#include <LiDIA/sort_vector.h>
#include <LiDIA/sort_vector.c>
#include <LiDIA/lidia_vector.h>
#endif

#if defined (BASE_VECTOR)
template class base_vector < TYPE >;
#endif

#if defined (FILE_VECTOR)
template class base_vector < TYPE >;
template class file_vector < TYPE >;
#endif

#if defined (MATH_VECTOR)
template class base_vector < TYPE >;
template class math_vector < TYPE >;
#endif 

#if defined (SORT_VECTOR)
template class base_vector < TYPE >;
template class sort_vector < TYPE >;
#endif

#if defined (LIDIA_VECTOR) 
template class base_vector < TYPE >;
template class file_vector < TYPE >;
template class math_vector < TYPE >;
template class sort_vector < TYPE >;
template class lidia_vector < TYPE >;
#endif












