#ifndef LIDIA_RANDOM_H
#define LIDIA_RANDOM_H

#include <LiDIA/config.h>
#include <LiDIA/iint.h>

#if (!defined(HAVE_RANDOM) || (defined(sun) && (defined(__GNUC__) || defined(__SUNPRO_CC))))
EXTERN_FUNCTION(void srandom, ( unsigned x ));
EXTERN_FUNCTION(char * initstate, ( unsigned seed, char *arg_state, int n ));
EXTERN_FUNCTION(char * setstate, ( char *arg_state ));
EXTERN_FUNCTION(long random, (_VOID_));
#else
#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif
#endif

#endif
