
#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)
#include <LiDIA:base_matrix.h>
#include <LiDIA:base_matrix.c>

#if defined (MATH_MATRIX) 
#include <LiDIA:math_matrix.h>
#include <LiDIA:math_matrix.c>
#endif

#else

#include <LiDIA/base_matrix.h>
#include <LiDIA/base_matrix.c>

#if defined (MATH_MATRIX) 
#include <LiDIA/math_matrix.h>
#include <LiDIA/math_matrix.c>
#endif

#endif

#if defined (BASE_MATRIX)
template class base_matrix < TYPE >;
#endif

#if defined (MATH_MATRIX)
template class base_matrix < TYPE >;
template class math_matrix < TYPE >;
#endif





