


#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)
#include <LiDIA:bigint.h>
#include <LiDIA:bigmod.h>
#include <LiDIA:bigrational.h>
#include <LiDIA:bigcomplex.h>

#include <LiDIA:base_dense_power_serie.h>
#include <LiDIA:base_dense_power_serie.c>


#else
#include <LiDIA/bigint.h>
#include <LiDIA/bigmod.h>
#include <LiDIA/bigrational.h>
#include <LiDIA/bigcomplex.h>

#include <LiDIA/base_dense_power_serie.h>
#include <LiDIA/base_dense_power_serie.c>
#endif


template class base_dense_power_serie< TYPE >;
