//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995, 1996 by the LiDIA Group
//
// File        : poly_error_msg.cc
// Author      : Stefan Neis (SN)
//               

/*
$Id
*/

const char *poly_error_msg[] =
{
  "(0) Error found in parameters !",
  "(1) Unknown input format for polynomials !",
  "(2) Parameters out of range !"
};
