//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995 by the LiDIA Group
//


#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)

#include <LiDIA:base_factor.h>

#else
  
#include <LiDIA/base_factor.h>

#endif



/****************************************************************
			class base_factor< T >
****************************************************************/

	/**
	** constructors, destructor
	**/
	
template< class T >
base_factor< T >::base_factor()
{
    set_prime_flag(not_prime);
}

template< class T >
base_factor < T >::base_factor(const base_factor < T > & x)
{
    rep = x.rep;
    set_prime_flag(x.prime_flag());
}

