//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995, 1996 by the LiDIA Group
//
// File        : gf_pol_instant.c
// Author      : Thomas Pfahler (TPf)
// Last change : TPf, initial version
//

#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)

#include <LiDIA:gf_polynomial.h>
#include <LiDIA:polynomial.c>

#else

#include <LiDIA/gf_polynomial.h>
#include <LiDIA/polynomial.c>

#endif

template class base_polynomial < gf_p_element >;
template class field_polynomial < gf_p_element >;


