! Integer Version 2.0, RD, 11.2.93	idigitvecmult.s	*/

#include "lidia_elf.h"

	.global	USCORE(DigitVecMult)
USCORE(DigitVecMult):
	tst	%o3
	ble	Lvma1
	mov	0, %g5		! carry
Lvma2:
	mov	%o2, %y
	andcc	%g0, %g0, %g3
	ld	[%o1], %g2	! *a
	nop

	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3

	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3

	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3

	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3

	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3

	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3

	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3

	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3
	mulscc	%g3, %g2, %g3

	mulscc	%g3, %g0, %g3		! 33rd iteration, prod high

	tst	%g2			! sign correction
	bge	Lvma3
	rd	%y, %g2			! prod low
	add	%g3, %o2, %g3		! sign correction
Lvma3:
	addcc	%g2, %g5, %g2		! carry + prod (low) -> result low
	addx	%g3, %g0, %g5		! result high -> carry
	st	%g2, [%o0]		! *res
	inc	4, %o1			! a++
	subcc	%o3, 1, %o3		! count--
	bg	Lvma2			! loop
	inc	4, %o0
Lvma1:
	retl
	mov	%g5, %o0		! return carry


