#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)

#include <LiDIA:gf_p_element.h>
//#include <LiDIA:gf_n_element.h>

#include <LiDIA:gf_element.h>
#include <LiDIA:gf_element.c>

#else

#include <LiDIA/gf_p_element.h>
//#include <LiDIA/gf_n_element.h>

#include <LiDIA/gf_element.h>
#include <LiDIA/gf_element.c>

#endif


template class gf_element < TYPE >;
template void element_power(TYPE & c, const TYPE & a, const bigint & e);


