



#if defined(HAVE_MAC_DIRS) || defined(__MWERKS__)
#include <LiDIA:bigint.h>
#include <LiDIA:bigrational.h>
#include <LiDIA:bigcomplex.h>

#define NO_PSR_BIGMOD

#include <LiDIA:coeff_sparse_power_serie.h>
#include <LiDIA:coeff_sparse_power_serie.c>


#include <LiDIA:base_sparse_power_serie.h>
#include <LiDIA:base_sparse_power_serie.c>

#include <LiDIA:sparse_power_serie.h>
#include <LiDIA:sparse_power_serie.c>


#else

#include <LiDIA/bigint.h>
#include <LiDIA/bigrational.h>
#include <LiDIA/bigcomplex.h>

#define NO_PSR_BIGMOD

#include <LiDIA/coeff_sparse_power_serie.h>
#include <LiDIA/coeff_sparse_power_serie.c>


#include <LiDIA/base_sparse_power_serie.h>
#include <LiDIA/base_sparse_power_serie.c>

#include <LiDIA/sparse_power_serie.h>
#include <LiDIA/sparse_power_serie.c>

#endif


template class                     spc< TYPE >;
template class base_sparse_power_serie< TYPE >;
template class      sparse_power_serie< TYPE >;


