#include "lidia_elf.h"

	.text
	.align 4
	.globl USCORE(DigitVecCsubto)
USCORE(DigitVecCsubto):
	pushl %ebx
	pushl %edi
	movl 12(%esp),%eax
	movl 16(%esp),%ebx
	movl 20(%esp),%ecx
	xorl %edi,%edi
	jecxz L3
	clc
L1:
	movl (%eax,%edi,4),%edx
	sbbl (%ebx,%edi,4),%edx
	movl %edx,(%eax,%edi,4)
	incl %edi
	loop L1 
	jnb L3
L2:
	subl $1,(%eax,%edi,4)
	incl %edi
	jb L2
L3: 
	popl %edi
	popl %ebx
	ret
