//
// LiDIA - a library for computational number theory
//   Copyright (c) 1994, 1995 by the LiDIA Group
//
// File        : sparse_matrix.h
// Author      : Patrick Theobald (PT)
// Last change : PT, Jun 16, 1996, initial version
//             

/*
$Id: sparse_matrix.h,v 1.2 1996/10/21 15:13:25 theobald Exp $
*/

#ifndef LIDIA_SPARSE_MATRIX_H
#define LIDIA_SPARSE_MATRIX_H

template < class T >
class sparse_matrix
{
protected:	
  
  lidia_size_t sparse_rows;
  lidia_size_t sparse_columns;
  
  lidia_size_t **index_array;

  lidia_size_t *value_counter;
};

#endif






