/* Integer Version 2.0, RD, 15.1.93	imem.h	*/

#ifndef _IMEM_H
#define _IMEM_H


#include "iint.h"

#if defined(__STDC__) || defined(__MWERKS__)
#include <stdlib.h>
#else
#include <malloc.h>
#endif

EXTERN_FUNCTION(DigitType * newDigitVec, (int *maxl));
EXTERN_FUNCTION(void delDigitVec, (DigitType *u, int maxl));

EXTERN_FUNCTION(Integer * _newInteger, (_VOID_));
EXTERN_FUNCTION(void _delInteger, (Integer *u));

EXTERN_FUNCTION(char *Imalloc, (int));	/* Systemunabhaengiges malloc */
EXTERN_FUNCTION(void Ifree, (char *)); /* Systemunabhaengiges free */
EXTERN_FUNCTION(void IprintStatistics, (_VOID_));


#endif
